""" DOM-to-text serialisers for PXTL implementations.
"""

__all__= [
  'serialisers', 'Serialiser',
  'XMLSerialiser', 'XHTMLSerialiser', 'HTMLSerialiser', 'TextSerialiser'
]

import string
from pxtl.constants import *


class Serialiser:
  """ Interface for serialisers. (This is going to become considerably more
      tricky with the optimised implementation.)
  """
  name= '(unknown serialiser)'
  def __init__(self, writer, encoding):
    if encoding is None:
      self.w= writer.write
    else:
      self.w= Encoder(writer, encoding).write
  def writeDocument(self, node, literals):
    self.writeChildren(node)
  def writeChildren(self, node):
    for child in node.childNodes:
      if child.nodeType==node.ELEMENT_NODE:
        self.writeElement(child)
      elif child.nodeType==node.TEXT_NODE:
        self.writeText(child)
      elif child.nodeType==node.ENTITY_REFERENCE_NODE:
        self.writeEntity(child)
      elif child.nodeType==node.COMMENT_NODE:
        self.writeComment(child)
      elif child.nodeType==node.CDATA_SECTION_NODE:
        self.writeCDATA(child)
      elif child.nodeType==node.PROCESSING_INSTRUCTION_NODE:
        self.writePI(child)
      elif child.nodeType==node.DOCUMENT_TYPE_NODE:
        self.writeDoctype(child)
  def writeElement(self, node):
    for i in range(node.attributes.length):
      self.writeAttribute(node.attributes.item(i))
    self.writeChildren(node)
  def writeAttribute(self, node): pass
  def writeText(self, node): pass
  def writeEntity(self, node): pass
  def writeComment(self, node): pass
  def writeCDATA(self, node): pass
  def writePI(self, node): pass
  def writeDoctype(self, node): pass


class XMLSerialiser(Serialiser):
  """ Serialiser for plain XML.
  """
  name= 'xml'
  def __init__(self, writer, encoding):
    Serialiser.__init__(self, writer, encoding)
    self.inattr= False
  def writeDocument(self, node, literals):
    self.literals= literals
    prolog= True
    try:
      prolog= None not in (
        node.xmlVersion, node.xmlEncoding, node.xmlStandalone
      )
    except AttributeError:
      prolog= False
    if prolog:
      if node.xmlVersion is not None:
        self.w('<?xml version="%s"' % node.xmlVersion)
      else:
        self.w('<?xml version="1.0"')
      if node.xmlEncoding is not None:
        self.w(' encoding="%s"' % node.xmlEncoding)
      if node.xmlStandalone:
        self.w(' standalone="yes"')
      self.w('?>')
    self.writeChildren(node)

  def writeElement(self, node):
    self.w('<')
    self.w(node.tagName)
    for i in range(node.attributes.length):
      self.writeAttribute(node.attributes.item(i))
    if (len(node.childNodes)==0):
      self.w('/>')
    else:
      self.w('>')
      self.writeChildren(node)
      self.w('</')
      self.w(node.tagName)
      self.w('>')

  def writeAttribute(self, node):
    self.w(' ')
    self.w(node.name)
    self.w('="')
    inattr= self.inattr
    self.inattr= True

    # Here is PXTL's single concession to dodgy DOM implementations. Only
    # use the child nodes of an attribute if they are present and reliable.
    # Otherwise (that's you, minidom) just use the value string.
    #
    child= node.firstChild
    if child is not None and (
      child.nodeType!=node.TEXT_NODE or hasattr(child, 'data')
    ):
      self.writeChildren(node)
    else:
      self.writeString(node.value)
    self.inattr= inattr
    self.w('"')

  def writeText(self, node):
    if node in self.literals:
      self.w(node.data)
    else:
      self.writeString(node.data)

  def writeEntity(self, node):
    self.w('&')
    self.w(node.nodeName)
    self.w(';')

  def writeComment(self, node):
    data= string.replace(node.data, '--', '- - ')
    self.w('<!--')
    self.w(data)
    if data[-1:]=='-':
      self.w(' ')
    self.w('-->')    

  def writeCDATA(self, node):
    first= True
    for data in string.split(node.data, ']]>'):
      if first:
        first= False
      else:
        self.w(']]>')
      self.w('<[CDATA[')
      self.w(data)
      self.w(']]>')

  def writePI(self, node):
    self.w('<?')
    self.w(node.target)
    self.w(' ')
    self.w(node.data)
    self.w('?>')

  def writeDoctype(self, node):
    self.w('<!DOCTYPE %s ' % node.nodeName)
    if node.publicId==None:
      self.w('SYSTEM')
    else:
      self.w('PUBLIC "')
      self.writeString(node.publicId)
      self.w('"')
    if node.systemId!=None:
      self.w(' "')
      self.writeString(node.systemId)
      self.w('"')
    if node.internalSubset not in (None, ''):
      self.w(' [%s]' % node.internalSubset)
    self.w('>')

  def writeString(self, s):
    r= string.replace
    if self.inattr:
      self.w( r(r(r(r(r(r(s,
        '&', '&amp;'), '<', '&lt;'), '"', '&quot;'), ']]>', ']]&gt;'),
        '\n', '&#10;'), '\t', '&#9;')
      )
    else:
      self.w(r(r(r(s,'&','&amp;'),'<','&lt;'),']]>',']]&gt;'))

  def getContentType(self, node):
    t= 'text/plain'
    if node.namespaceURI in (HTNS, None) and node.localName in HTML_CDATA:
      if node.localName=='style':
        t= 'text/css'
      elif node.localName=='script':
        t= 'text/javascript'
      if node.hasAttributeNS(None, 'type'):
        t= node.getAttributeNS(None, 'type')
    return t


class XHTMLSerialiser(XMLSerialiser):
  """ XML serialiser whose XHTML output will be compatible with legacy HTML,
      as described in XHTML 1.0 appendix C.
  """
  name= 'xhtml'
  def writeElement(self, node):
    self.w('<')
    self.w(node.tagName)
    for i in range(node.attributes.length):
      attr= node.attributes.item(i)
      self.writeAttribute(attr)
      if (
        node.namespaceURI in [HTNS, None] and
        attr.namespaceURI==XMNS and attr.localName=='lang' and
        not node.hasAttributeNS(node.namespaceURI, 'lang')
      ):
        hattr= node.ownerDocument.createAttributeNS(node.namespaceURI, 'lang')
        hattr.value= attr.value
        self.writeAttribute(hattr)
    if node.namespaceURI in [HTNS, None] and node.localName in HTML_EMPTY:
      self.w(' />')
    else:
      self.w('>')
      self.writeChildren(node)
      self.w('</')
      self.w(node.tagName)
      self.w('>')

  def writeCDATA(self, node):
    t= self.getContentType(node.parentNode)
    if t=='text/css':
      self.w('<!--/*--><![CDATA[/*><!--*/')
      self.w(node.data)
      self.w('/*]]>*/-->')
    elif t in ('text/javascript', 'text/ecmascript', 'text/jscript'):
      self.w('<!--/*--><![CDATA[/*><!--*/')
      self.w(node.data)
      self.w('//]]>-->')
    elif t=='text/vbscript':
      self.w('<!--\'--><![CDATA[\'><!--\n')
      self.w(node.data)
      self.w('\']]>-->')
    else:
      self.writeString(node.data)


class HTMLSerialiser(XMLSerialiser):
  name= 'html'
  def writeElement(self, node):
    self.w('<')
    self.w(node.tagName)
    for i in range(node.attributes.length):
      attr= node.attributes.item(i)
      if (
        node.namespaceURI in [HTNS, None] and
        attr.namespaceURI==XMNS and attr.localName=='lang' and
        not node.hasAttributeNS(node.namespaceURI, 'lang')
      ):
        hattr= node.ownerDocument.createAttributeNS(node.namespaceURI, 'lang')
        hattr.value= attr.value
        self.writeAttribute(hattr)
      else:
        self.writeAttribute(attr)
    self.w('>')
    self.writeChildren(node)
    if not (node.namespaceURI in [HTNS, None] and node.localName in HTML_EMPTY):
      self.w('</')
      self.w(node.tagName)
      self.w('>')

  def writeAttribute(self, node):
    if node.namespaceURI in (NSNS, XMNS):
      return
    self.w(' ')
    if (
      node.namespaceURI in (HTNS, None) and
      node.localName in HTML_MINIMISABLE and
      string.lower(node.name)==string.lower(node.value)
    ):
      self.w(node.value)
    else:
      self.w(node.name)
      self.w('="')
      self.w(node.value)
      self.w('"')

  def writeCDATA(self, node):
    t= self.getContentType(node.parentNode)
    if t=='text/css':
      self.w('<!--')
      self.w(node.data)
      self.w('-->')
    elif t in ('text/javascript', 'text/ecmascript', 'text/jscript'):
      self.w('<!--')
      self.w(node.data)
      self.w('//-->')
    elif t=='text/vbscript':
      self.w('<!--')
      self.w(node.data)
      self.w('\'-->')
    else:
      self.writeString(node.data)

class TextSerialiser(Serialiser):
  """ Plain-text serialiser. Never outputs any markup. All text is sent to
      write() unescaped.
  """
  name= 'text'
  def writeText(self, node):
    self.w(node.data)
  def writeCDATA(self, node):
    self.w(node.data)
  def writeEntity(self, node):
    self.writeChildren(node)


# HTML element info constants for xhtml and html serialisers
#
HTML_EMPTY= [
  'area', 'base', 'basefont', 'br', 'col', 'frame', 'hr', 'img',
  'input', 'isindex', 'link', 'meta', 'param'
]

HTML_CDATA= ['style', 'script']

HTML_MINIMISABLE= [
  'checked', 'compact', 'declare', 'defer', 'disabled', 'ismap', 'multiple',
  'nohref', 'noresize', 'noshade', 'nowrap', 'readonly', 'selected'
]

# Create mapping of serialiser names to their classes.
#
serialisers= {}
for s in [XMLSerialiser, XHTMLSerialiser, HTMLSerialiser, TextSerialiser]:
  serialisers[s.name]= s
del s

# Pipe class to encode unicode to strings before sending to an output stream.
# In Python 1.6, plain strings have no .encode method, so these have to be
# turned into Unicode first.
#
class Encoder:
  def __init__(self, writer, encoding):
    self.writer= writer
    self.encoding= encoding
  def write(self, s):
    if not isinstance(s, UnicodeType) and not hasattr(s, 'encode'):
      s= unicode(s, 'us-ascii')
    self.writer.write(s.encode(self.encoding))
